import sys
from os import environ
from .deployer import Executor
import pathlib

import yaml
from botocore.session import Session


INPUT_FILE_ENV = 'DEPLOY_OUTPUT_FILE' # todo file output


def execute_ado_cli():

    expected_env_vars = { INPUT_FILE_ENV }
    missing_vars = []
    for env_var in expected_env_vars:
        if not env_var in environ:
            missing_vars.append(env_var)
    if missing_vars:
        raise Exception(f'Missing environment variables: {missing_vars}')

    cs_result_file_path = environ[INPUT_FILE_ENV]

    cs_result_file_path = pathlib.Path(cs_result_file_path)
    if not cs_result_file_path.exists():
        raise Exception('Stack config file does not exists.')
    cs_result_yaml = cs_result_file_path.read_text()
    cs_result = yaml.safe_load(str(cs_result_yaml))

    stack = execute(
        aws_region=cs_result['stack_region'],
        stack_id=cs_result['changeset']['StackId'],
        changeset_id=cs_result['changeset_id'],
        changeset_type=cs_result['changeset_type'],
    )

    print(yaml.safe_dump(stack))

    if 'Outputs' in stack:
        for output in stack['Outputs']:
            var_name = output['OutputKey']
            var_value = output['OutputValue']
            print(f'##vso[task.setvariable variable={var_name}]{var_value}')


def execute(aws_region, stack_id, changeset_id, changeset_type):

    boto_session = Session()

    cfn_client = boto_session.create_client('cloudformation', region_name=aws_region)

    executor = Executor(cfn_client)
    executor.execute_changeset(changeset_id)
    executor.wait_for_execute(stack_name=stack_id, changeset_type=changeset_type)
    stack = executor.describe_stack(stack_id)
    return stack


if __name__ == '__main__':
    execute_ado_cli()


import typer
import os
import sys
import yaml
from pathlib import Path
from os import environ
from .deploy import deploy_py, execute

app = typer.Typer()


@app.callback()
def callback():
    """
    CloudFormation Control CLI
    """

@app.command("plan")
def plan_cloudformation(
    # Required
    changeset_name:             str = typer.Option(..., "--name", envvar="CFNCTL_CHANGESET_NAME", help="CloudFormation change set name."),
    changeset_description:      str = typer.Option(..., "--description", envvar="CFNCTL_CHANGESET_DESCRIPTION", help="CloudFormation change set description."),
    template_path:              Path = typer.Option(..., "--template", resolve_path=True, envvar="CFNCTL_TEMPLATE_PATH", help="Relative or absolute filepath pointing to CloudFormation template file."),
    config_path:                Path = typer.Option(..., "--config", resolve_path=True, envvar="CFNCTL_STACK_CONFIG_PATH", help="Relative or absolute filepath pointing to CloudFormation template config file."),
    plan_output_path:           Path = typer.Option(..., "--output", envvar="CFNCTL_PLAN_OUTPUT_PATH", help="Relative or absolute filepath for resulting change set file."),
    # Optional
    stack_id_var_name:          str = typer.Option("", "--export-stack", envvar="CFNCTL_STACK_ID_EXPORT_VAR", help="Export the Stack ID as an Azure DevOps pipeline variable with this name."),
    changeset_id_var_name:      str = typer.Option("", "--export-changeset", envvar="CFNCTL_CHANGESET_ID_EXPORT_VAR", help="Export the change set ID as an Azure DevOps pipeline variable with this name."),
    export_variables:           bool = typer.Option(False, "--export-outputs", envvar="CFNCTL_EXPORT_STACK_OUTPUT", help="Exports the stack output as Azure DevOps pipeline variables."),
    auto_apply:                 bool = typer.Option(False, "--auto-apply", envvar="CFNCTL_AUTO_APPLY", help="If set to true, changeset will be created then executed"),
):
    """
    Create a new CloudFormation stack / change set plan based on provided template and config file. 
    """
    # Get config file and convert yaml to python dict
    conf_str = config_path.read_text()
    conf_dict = yaml.safe_load(str(conf_str))

    # Check output filepath
    out_file_path = Path(plan_output_path)
    out_file_folder = out_file_path.parent
    out_file_path = str(out_file_path)
    os.makedirs(out_file_folder, exist_ok=True)

    result = deploy_py(conf_dict, changeset_name, changeset_description, template_path)
    result_yaml = yaml.safe_dump(result)
    print(result_yaml)

    f_out = open(out_file_path, 'w')
    f_out.write(result_yaml)
    f_out.close()

    # If auto apply set to True, execute CloudFormation change set using variables from Plan as input.
    if auto_apply:
        apply_cloudformation(changeset_name, changeset_description, template_path, config_path, plan_output_path, stack_id_var_name, changeset_id_var_name, export_variables, auto_apply)

    # Sets variables in Azure DevOps pipeline if parameters are set
    if changeset_id_var_name and not auto_apply:
        changeset_id = result['changeset_id']
        print(f'##vso[task.setvariable variable={changeset_id_var_name}]{changeset_id}')
    if stack_id_var_name and not auto_apply:
        stack_id = result['changeset']['StackId']
        print(f'##vso[task.setvariable variable={stack_id_var_name}]{stack_id}')

    return result['exit_code']

@app.command("apply")
def apply_cloudformation(
    # Required
    changeset_name:             str = typer.Option(..., "--name", envvar="CFNCTL_CHANGESET_NAME", help="CloudFormation change set name."),
    changeset_description:      str = typer.Option(..., "--description", envvar="CFNCTL_CHANGESET_DESCRIPTION", help="CloudFormation change set description."),
    template_path:              Path = typer.Option(..., "--template", resolve_path=True, envvar="CFNCTL_TEMPLATE_PATH", help="Relative or absolute filepath pointing to CloudFormation template file."),
    config_path:                Path = typer.Option(..., "--config", resolve_path=True, envvar="CFNCTL_STACK_CONFIG_PATH", help="Relative or absolute filepath pointing to CloudFormation template config file."),
    plan_output_path:           Path = typer.Option(..., "--output", envvar="CFNCTL_PLAN_OUTPUT_PATH", help="Relative or absolute filepath for resulting change set file."),
    # Optional
    stack_id_var_name:          str = typer.Option("", "--export-stack", envvar="CFNCTL_STACK_ID_EXPORT_VAR", help="Export the Stack ID as an Azure DevOps pipeline variable with this name."),
    changeset_id_var_name:      str = typer.Option("", "--export-changeset", envvar="CFNCTL_CHANGESET_ID_EXPORT_VAR", help="Export the change set ID as an Azure DevOps pipeline variable with this name."),
    export_variables:           bool = typer.Option(False, "--export-outputs", envvar="CFNCTL_EXPORT_STACK_OUTPUT", help="Exports the stack output as Azure DevOps pipeline variables."),
    auto_apply:                 bool = typer.Option(False, "--auto-apply", envvar="CFNCTL_AUTO_APPLY", help="If set to true, changeset will be created then executed"),
):
    """
    Apply a CloudFormation stack / change set based on provided template and config file. 
    """

    cs_result_file_path = plan_output_path

    cs_result_file_path = Path(cs_result_file_path)
    if not cs_result_file_path.exists():
        raise Exception('Stack config file does not exists.')
    cs_result_yaml = cs_result_file_path.read_text()
    cs_result = yaml.safe_load(str(cs_result_yaml))

    stack_id = cs_result['changeset']['StackId']

    stack = execute(
        aws_region=cs_result['stack_region'],
        stack_id=stack_id,
        changeset_id=cs_result['changeset_id'],
        changeset_type=cs_result['changeset_type'],
    )

    print(yaml.safe_dump(stack))

    if export_variables and 'Outputs' in stack:
        for output in stack['Outputs']:
            var_name = output['OutputKey']
            var_value = output['OutputValue']
            print(f'##vso[task.setvariable variable={var_name}]{var_value}')

    if changeset_id_var_name:
        changeset_id = cs_result['changeset_id']
        print(f'##vso[task.setvariable variable={changeset_id_var_name}]{changeset_id}')
    if stack_id_var_name:
        print(f'##vso[task.setvariable variable={stack_id_var_name}]{stack_id}')
    
    return stack


if __name__ == "__main__":
    typer.run(app)


# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# Modified version of this file:
# https://github.com/aws/aws-cli/blob/develop/awscli/customizations/cloudformation/deploy.py

import os
from os import environ
import sys
import logging
import pathlib

import yaml
from botocore.session import Session

from .exceptions import *
from .deployer import Deployer, Executor
from .yamlhelper import yaml_parse


LOG = logging.getLogger(__name__)


CAPABILITIES = {'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM', 'CAPABILITY_AUTO_EXPAND'}

MSG_EXECUTE_SUCCESS = "Successfully created/updated stack - {stack_name}\n"

MSG_NO_EXECUTE_CHANGESET = \
    ("Changeset created successfully. Run the following command to "
        "review changes:"
        "\n"
        "aws cloudformation describe-change-set --change-set-name "
        "{changeset_id}"
        "\n")


def deploy_py(conf_dict, changeset_name, changeset_description, template_path,
            capabilities=CAPABILITIES, exeute_changeset=False,
            fail_on_empty_changeset=False):
    capabilities = list(capabilities)
    stack_name = conf_dict['StackName']
    aws_region = conf_dict['AwsRegion']
    parameter_overrides = conf_dict['StackParameters']
    tags_dict = conf_dict['StackTags']

    boto_session = Session()
    
    # ---- from _run_main

    cloudformation_client = \
        boto_session.create_client('cloudformation', region_name=aws_region)
        
    if not os.path.isfile(template_path):
        raise InvalidTemplatePathError(
                template_path=template_path)

    # Parse parameters
    with open(template_path, "r") as handle:
        template_str = handle.read()

    tags = [{"Key": key, "Value": value}
                for key, value in tags_dict.items()]
    
    template_dict = yaml_parse(template_str)

    parameters = merge_parameters(template_dict, parameter_overrides)
    
    template_size = os.path.getsize(template_path)
    if template_size > 51200:
        raise DeployBucketRequiredError()

    deployer = Deployer(cloudformation_client, changeset_name, changeset_description)
    return deploy(deployer, stack_name, template_str, aws_region,
                    parameters, capabilities,
                    exeute_changeset, None, None,
                    tags,
                    fail_on_empty_changeset)

def deploy(deployer, stack_name, template_str, aws_region,
            parameters, capabilities, execute_changeset, role_arn,
            notification_arns, tags,
            fail_on_empty_changeset=True):
    try:
        result = deployer.create_and_wait_for_changeset(
            stack_name=stack_name,
            cfn_template=template_str,
            parameter_values=parameters,
            capabilities=capabilities,
            role_arn=role_arn,
            notification_arns=notification_arns,
            tags=tags
        )
    except ChangeEmptyError as ex:
        if fail_on_empty_changeset:
            raise
        changeset = ex.kwargs['changeset']
        deployer.delete_changeset(changeset['ChangeSetId'])
        return {
            'exit_code': 0,
            'status': 'NO_CHANGES',
            'message': 'The ChangeSet contained no changes and has been deleted.',
            'stack_region': aws_region,
            'stack_name': stack_name,
            'changeset_name': deployer.changeset_name,
            'changeset_type': 'NO_CHANGES',
            'changeset_id': changeset['ChangeSetId'],
            'changeset': changeset,
        }

    if execute_changeset:
        deployer.execute_changeset(result.changeset_id, stack_name)
        deployer.wait_for_execute(stack_name, result.changeset_type)
        sys.stdout.write(MSG_EXECUTE_SUCCESS.format(
                stack_name=stack_name))
        status = 'UPDATES_APPLIED'
        message = 'The ChangeSet was executed successfully.'
    else:
        sys.stdout.write(MSG_NO_EXECUTE_CHANGESET.format(
                changeset_id=result.changeset_id))
        status = 'UPDATES_PENDING_APPROVAL'
        message = 'The ChangeSet is pending execution.'

    sys.stdout.flush()
    return {
        'exit_code': 0,
        'status': status,
        'message': message,
        'stack_region': aws_region,
        'stack_name': stack_name,
        'changeset_name': deployer.changeset_name,
        'changeset_type': result.changeset_type,
        'changeset_id': result.changeset_id,
        'changeset': result.changeset,
    }

def merge_parameters(template_dict, parameter_overrides):
    """
    CloudFormation CreateChangeset requires a value for every parameter
    from the template, either specifying a new value or use previous value.
    For convenience, this method will accept new parameter values and
    generates a dict of all parameters in a format that ChangeSet API
    will accept

    :param parameter_overrides:
    :return:
    """
    parameter_values = []

    if not isinstance(template_dict.get("Parameters", None), dict):
        return parameter_values

    for key, value in template_dict["Parameters"].items():

        obj = {
            "ParameterKey": key
        }

        if key in parameter_overrides:
            obj["ParameterValue"] = parameter_overrides[key]
        else:
            obj["UsePreviousValue"] = True

        parameter_values.append(obj)

    return parameter_values


# def execute_ado_cli():

#     expected_env_vars = { env.OUTPUT_FILE_ENV }
#     missing_vars = []
#     for env_var in expected_env_vars:
#         if not env_var in environ:
#             missing_vars.append(env_var)
#     if missing_vars:
#         raise Exception(f'Missing environment variables: {missing_vars}')

#     cs_result_file_path = environ[env.OUTPUT_FILE_ENV]
    
#     export_stack_outputs = env.get_flag(env.OUTPUT_AS_VARIABLES_ENV, False)
#     changeset_id_var_name = env.get_optional_var(env.CHANGESET_ID_VAR_NAME_ENV)
#     stak_id_var_name = env.get_optional_var(env.STACK_ID_VAR_NAME_ENV)

#     cs_result_file_path = pathlib.Path(cs_result_file_path)
#     if not cs_result_file_path.exists():
#         raise Exception('Stack config file does not exists.')
#     cs_result_yaml = cs_result_file_path.read_text()
#     cs_result = yaml.safe_load(str(cs_result_yaml))

#     stack_id = cs_result['changeset']['StackId']

#     stack = execute(
#         aws_region=cs_result['stack_region'],
#         stack_id=stack_id,
#         changeset_id=cs_result['changeset_id'],
#         changeset_type=cs_result['changeset_type'],
#     )

#     print(yaml.safe_dump(stack))

#     if export_stack_outputs and 'Outputs' in stack:
#         for output in stack['Outputs']:
#             var_name = output['OutputKey']
#             var_value = output['OutputValue']
#             print(f'##vso[task.setvariable variable={var_name}]{var_value}')

#     if changeset_id_var_name:
#         changeset_id = cs_result['changeset_id']
#         print(f'##vso[task.setvariable variable={changeset_id_var_name}]{changeset_id}')
#     if stak_id_var_name:
#         print(f'##vso[task.setvariable variable={stak_id_var_name}]{stack_id}')
    
#     return stack


def execute(aws_region, stack_id, changeset_id, changeset_type):

    boto_session = Session()

    cfn_client = boto_session.create_client('cloudformation', region_name=aws_region)

    executor = Executor(cfn_client)
    executor.execute_changeset(changeset_id)
    executor.wait_for_execute(stack_name=stack_id, changeset_type=changeset_type)
    stack = executor.describe_stack(stack_id)
    return stack


# if __name__ == '__main__':

#     options = ['deploy', 'execute']
#     if len(sys.argv) < 2:
#         raise Exception(f'Must give 1 of 2 arguments: {options}')
#     option = sys.argv[1]
    
#     if option == 'deploy':
#         deploy_ado_cli()
#     elif option == 'execute':
#         execute_ado_cli()


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nn_cfnctl']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'boto3>=1.20.39,<2.0.0', 'typer[all]>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['cfnctl = nn_cfnctl.main:app']}

setup_kwargs = {
    'name': 'nn-cfnctl',
    'version': '0.2.0',
    'description': 'CLI Tool for deploying CloudFormation using a YAML based config file for stack and parameter settings.',
    'long_description': '# CloudFormation Control\n\n<!-- ABOUT THE PROJECT -->\n## About The Project\n\nCloud Formation Control (cfnctl) is a commandline utility based of the AWS CLI Deploy features.\nOn top of existing deployment capabilities it adds the ability to specify a configuration file to easily supply the following parameters to your CloudFormation deployment:\n\n* StackName\n* AwsRegion\n* StackParameters\n* StackTags\n\nSupplying your CloudFormation template with a config containing the above parameters, and a set of commandline arguments or environment variables, it makes it easier to deploy CloudFormation templates both from your local development machine as well as a pipeline using the same pipelines and configurations.\n\n\n### Built With\n\nThis project is build using the following tools and frameworks:\n\n* [Python](https://www.python.org/)\n* [Boto3](https://github.com/boto/boto3)\n* [Typer](https://github.com/tiangolo/typer)\n* [Poetry](https://github.com/python-poetry/poetry)\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nIn order to get started with developing and contributing to this project, follow the steps below.\n\n### Prerequisites\n\n#### pipx (Optional)\n\nThis project is build and managed by poetry, the easiest way to install poetry in an isolated environment is using [pipx][pipx-install].\n\n```sh\npython -m pip install --user pipx\npython -m pipx ensurepath\n```\n\n#### Poetry\n\nIf you installed pipx, you can simply install poetry using the command below:\n\n```sh\npipx install poetry\n```\n\nIf you did not install pipx, see [Poetry documentation][poetry-install] for installation method for your system.\n\n### Installation\n\n1. Clone the repository\n   ```sh\n   git clone https://novonordiskit@dev.azure.com/novonordiskit/GITO-HS/_git/aws-cfnctl\n   ```\n2. Change directory into the cloned repository\n   ```sh\n   cd aws-cfnctl\n   ```\n3. Install dependencies into a virtual environment with poetry\n   ```sh\n   poetry install\n   ```\n3. Activate virtual environment with poetry\n   ```sh\n   poetry shell\n   ```\n\n<!-- USAGE EXAMPLES -->\n## Usage\n\nWhen installed somewhere on the system included in the systems PATH, the application can be called with `cfnctl`.  \nTo see all available options for the CLI run:\n```sh\ncfnctl --help  \n```\nThe `cfnctl` command has two primary subcommands:\n* `cfnctl plan`\n* `cfnctl apply`\n\n### cfnctl plan\n\n`cfnctl plan` creates a new stack plan, or if the stack already exists, a change set plan for AWS CloudFormation. \nThe resulting plan can then be used with `cfnctl apply` to execute the stack / change set and provision it in AWS.  \n\nIf you wish to execute the plan immediately, set the `--auto-apply` commandline flag or use the `CFNCTL_AUTO_APPLY = $TRUE`.\n\nThe `cfnctl plan` command requires the following options:\n* --name "CloudFormation change set name."\n* --description "CloudFormation change set description."\n* --template "Relative or absolute filepath pointing to CloudFormation template file."\n* --config "Relative or absolute filepath pointing to CloudFormation template config file."\n* --output "Relative or absolute filepath for resulting change set file."\n\nThe following settings can be set optionally:\n* --export-stack "Export the Stack ID as an Azure DevOps pipeline variable with this name."\n* --export-changeset "Export the change set ID as an Azure DevOps pipeline variable with this name."\n* --export-outputs "Exports the stack output as Azure DevOps pipeline variables."\n* --auto-apply "If set to true, changeset will be created then executed"\n\n### cfnctl apply\n\n`cfnctl apply` executes a stack plan, or if the stack already exists, executes a change set with AWS CloudFormation. \nThe command requires a plan file created with the `cfnctl plan` command.\n\n\nThe `cfnctl plan` command requires the following options:\n* --name "CloudFormation change set name."\n* --description "CloudFormation change set description."\n* --template "Relative or absolute filepath pointing to CloudFormation template file."\n* --config "Relative or absolute filepath pointing to CloudFormation template config file."\n* --output "Relative or absolute filepath for incoming change set file."\n\nThe following settings can be set optionally:\n* --export-stack "Export the Stack ID as an Azure DevOps pipeline variable with this name."\n* --export-changeset "Export the change set ID as an Azure DevOps pipeline variable with this name."\n* --export-outputs "Exports the stack output as Azure DevOps pipeline variables."\n\n### Environment variables\n\nInstead of using commandline parameters, it is possible to configure the `cfnctl plan` and `cfnctl apply` commands using environment variables.\n\n#### PowerShell environment variables\n\nSetting environment variables in PowerShell:\n\n```powershell\n$env:CFNCTL_CHANGESET_NAME = "MyChangeSet"\n$env:CFNCTL_CHANGESET_DESCRIPTION = "Change Set for my stack"\n$env:CFNCTL_TEMPLATE_PATH = "my-cfn-template-file.yml"\n$env:CFNCTL_STACK_CONFIG_PATH = "my-cfnctl-config-file.yml"\n$env:CFNCTL_PLAN_OUTPUT_PATH = "cfn-output-file.yml"\n\n# Optional\n$env:CFNCTL_STACK_ID_EXPORT_VAR = "CFNCTL_STACK_ID"\n$env:CFNCTL_CHANGESET_ID_EXPORT_VAR = "CFNCTL_CHANGESET_ID"\n$env:CFNCTL_EXPORT_STACK_OUTPUT = $TRUE\n$env:CFNCTL_AUTO_APPLY = $TRUE\n```\n\nYou can validate the values is set correctly in your shell with:\n```powershell\nGet-Item -Path Env:* | Where-Object {$_.name -like "CFNCTL_*"}\n```\n\n#### Bash environment variables\n\nSetting Environment variables in Bash:\n```sh\nexport CFNCTL_CHANGESET_NAME = "MyChangeSet"\nexport CFNCTL_CHANGESET_DESCRIPTION = "Change Set for my stack"\nexport CFNCTL_TEMPLATE_PATH = "my-cfn-template-file.yml"\nexport CFNCTL_STACK_CONFIG_PATH = "my-cfnctl-config-file.yml"\nexport CFNCTL_PLAN_OUTPUT_PATH = "cfn-output-file.yml"\n\n# Optional\nexport CFNCTL_STACK_ID_EXPORT_VAR = "CFNCTL_STACK_ID"\nexport CFNCTL_CHANGESET_ID_EXPORT_VAR = "CFNCTL_CHANGESET_ID"\nexport CFNCTL_EXPORT_STACK_OUTPUT = true\nexport CFNCTL_AUTO_APPLY = true\n```\n\nYou can validate the values is set correctly in your shell with:\n```sh\nenv | grep \'CFNCTL_\'\n```\n\n\n<!-- MARKDOWN LINKS & IMAGES -->\n[pipx-install]: https://pypa.github.io/pipx/installation/\n[poetry-install]: https://python-poetry.org/docs/master/#installing-with-the-official-installer\n',
    'author': 'Henrik Stanley Mortensen',
    'author_email': 'hsmx@novonordisk.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
